@@ -1,6 +1,6 @@
 /******************************************************************************
 * Copyright (c) 2011-2021, NVIDIA CORPORATION.  All rights reserved.
-* Modifications Copyright (c) 2021, Advanced Micro Devices, Inc.  All rights reserved.
+* Modifications Copyright (c) 2021-2025, Advanced Micro Devices, Inc.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
@@ -29,12 +29,14 @@
 #ifndef HIPCUB_ROCPRIM_BLOCK_BLOCK_MERGE_SORT_HPP_
 #define HIPCUB_ROCPRIM_BLOCK_BLOCK_MERGE_SORT_HPP_
 
+#include "../../../config.hpp"
+
 #include "../thread/thread_sort.hpp"
 #include "../util_math.hpp"
 #include "../util_type.hpp"
 
-#include <rocprim/detail/various.hpp>
-#include <rocprim/functional.hpp>
+#include <rocprim/detail/various.hpp> // IWYU pragma: export
+#include <rocprim/functional.hpp> // IWYU pragma: export
 
 BEGIN_HIPCUB_NAMESPACE
 
@@ -170,7 +172,8 @@ template <typename KeyT,
           typename ValueT,
           int NUM_THREADS,
           int ITEMS_PER_THREAD,
-          typename SynchronizationPolicy>
+          typename SynchronizationPolicy,
+          bool WARP_SORT = false>
 class BlockMergeSortStrategy
 {
   static_assert(PowerOfTwo<NUM_THREADS>::VALUE,
@@ -385,7 +388,7 @@ public:
       KeyT max_key = oob_default;
 
       #pragma unroll
-      for (int item = 1; item < ITEMS_PER_THREAD; ++item)
+      for (int item = WARP_SORT ? 1 : 0; item < ITEMS_PER_THREAD; ++item)
       {
         if (ITEMS_PER_THREAD * static_cast<int>(linear_tid) + item < valid_items)
         {
@@ -437,14 +440,16 @@ public:
 
       int thread_idx_in_thread_group_being_merged = mask & linear_tid;
 
+      const int COMPARE_NUM = WARP_SORT ? valid_items : ITEMS_PER_THREAD * blockDim.x;
+
       int diag =
-        (::rocprim::min)(valid_items,
+        (::rocprim::min)(COMPARE_NUM,
                    ITEMS_PER_THREAD * thread_idx_in_thread_group_being_merged);
 
-      int keys1_beg = (::rocprim::min)(valid_items, start);
-      int keys1_end = (::rocprim::min)(valid_items, keys1_beg + size);
+      int keys1_beg = (::rocprim::min)(COMPARE_NUM, start);
+      int keys1_end = (::rocprim::min)(COMPARE_NUM, keys1_beg + size);
       int keys2_beg = keys1_end;
-      int keys2_end = (::rocprim::min)(valid_items, keys2_beg + size);
+      int keys2_end = (::rocprim::min)(COMPARE_NUM, keys2_beg + size);
 
       int keys1_count = keys1_end - keys1_beg;
       int keys2_count = keys2_end - keys2_beg;
@@ -669,7 +674,6 @@ private:
   }
 };
 
-
 /**
  * @brief The BlockMergeSort class provides methods for sorting items
  *        partitioned across a CUDA thread block using a merge sorting method.
@@ -722,7 +726,7 @@ private:
  * __global__ void ExampleKernel(...)
  * {
  *     // Specialize BlockMergeSort for a 1D block of 128 threads owning 4 integer items each
- *     typedef hipcub::BlockMergeSort<int, 128, 4> BlockMergeSort;
+ *     using BlockMergeSort = hipcub::BlockMergeSort<int, 128, 4>;
  *
  *     // Allocate shared memory for BlockMergeSort
  *     __shared__ typename BlockMergeSort::TempStorage temp_storage_shuffle;
@@ -795,7 +799,7 @@ public:
 private:
   HIPCUB_DEVICE __forceinline__ void SyncImplementation() const
   {
-    CTA_SYNC();
+      __syncthreads();
   }
 
   friend BlockMergeSortStrategyT;
