@@ -1,7 +1,7 @@
 /******************************************************************************
  * Copyright (c) 2010-2011, Duane Merrill.  All rights reserved.
  * Copyright (c) 2011-2018, NVIDIA CORPORATION.  All rights reserved.
- * Modifications Copyright (c) 2021-2023, Advanced Micro Devices, Inc.  All rights reserved.
+ * Modifications Copyright (c) 2021-2025, Advanced Micro Devices, Inc.  All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions are met:
@@ -30,17 +30,21 @@
 #ifndef HIPCUB_ROCPRIM_UTIL_TYPE_HPP_
 #define HIPCUB_ROCPRIM_UTIL_TYPE_HPP_
 
+#include "../../config.hpp"
+#include "../../util_deprecated.hpp"
+
+#include <rocprim/detail/various.hpp> // IWYU pragma: export
+#include <rocprim/type_traits.hpp> // IWYU pragma: export
+#include <rocprim/types/future_value.hpp> // IWYU pragma: export
+
+#include <hip/hip_bfloat16.h>
+#include <hip/hip_fp16.h>
+#include <hip/hip_runtime.h>
+#include <hip/hip_vector_types.h>
+
 #include <limits>
 #include <type_traits>
 
-#include "../../config.hpp"
-
-#include <rocprim/detail/various.hpp>
-#include <rocprim/types/future_value.hpp>
-
-#include <hip/hip_fp16.h>
-#include <hip/hip_bfloat16.h>
-
 BEGIN_HIPCUB_NAMESPACE
 
 #ifndef DOXYGEN_SHOULD_SKIP_THIS    // Do not document
@@ -50,19 +54,7 @@ using NullType = ::rocprim::empty_type;
 #endif
 
 #ifndef HIPCUB_IS_INT128_ENABLED
-    #if defined(__HIPCC_RTC__)
-        #if defined(__HIPCC_RTC_INT128__)
-            #define HIPCUB_IS_INT128_ENABLED 1
-        #endif // !defined(__HIPCC_RTC_INT128__)
-    #else // !defined(__HIPCC_RTC__)
-        #if HIP_VERSION >= 50400000
-            #if(HIPCUB_HOST_COMPILER == HIPCUB_HOST_COMPILER_GCC)                         \
-                || (HIPCUB_HOST_COMPILER == HIPCUB_HOST_COMPILER_CLANG) || defined(__ICC) \
-                || defined(__GNUC__) || defined(__clang__)
-                #define HIPCUB_IS_INT128_ENABLED 1
-            #endif // GCC || CLANG || ICC
-        #endif // VER >= 5.4
-    #endif // !defined(__HIPCC_RTC__)
+    #define HIPCUB_IS_INT128_ENABLED 1
 #endif // !defined(HIPCUB_IS_INT128_ENABLED)
 
 template<bool B, typename T, typename F> struct
@@ -155,8 +147,8 @@ struct DoubleBuffer
     }
 };
 
-template <int A>
-struct Int2Type
+template<int A>
+struct HIPCUB_DEPRECATED_BECAUSE("Use ::std::integral_constant instead") Int2Type
 {
     enum {VALUE = A};
 };
@@ -204,9 +196,12 @@ using is_integral_or_enum =
 
 }
 
-template <typename NumeratorT, typename DenominatorT>
+// CUB deprecated this API, and suggests to use `::cuda::ceil_div` instead,
+// which is implemented in file `libcudacxx/include/cuda/__cmath/ceil_div.h`.
+template<typename NumeratorT, typename DenominatorT>
+HIPCUB_DEPRECATED_BECAUSE("Use hip::ceil_div instead from 'libhipcxx'")
 HIPCUB_HOST_DEVICE __forceinline__ constexpr NumeratorT
-DivideAndRoundUp(NumeratorT n, DenominatorT d)
+    DivideAndRoundUp(NumeratorT n, DenominatorT d)
 {
   static_assert(hipcub::detail::is_integral_or_enum<NumeratorT>::value &&
                 hipcub::detail::is_integral_or_enum<DenominatorT>::value,
@@ -239,16 +234,23 @@ struct AlignBytes
     };
 
     /// The "truly aligned" type
-    typedef T Type;
+    using Type = T;
 };
 
 // Specializations where host C++ compilers (e.g., 32-bit Windows) may disagree
 // with device C++ compilers (EDG) on types passed as template parameters through
 // kernel functions
 
-#define __HIPCUB_ALIGN_BYTES(t, b)         \
-    template <> struct AlignBytes<t>    \
-    { enum { ALIGN_BYTES = b }; typedef __align__(b) t Type; };
+    #define __HIPCUB_ALIGN_BYTES(t, b)                  \
+        template<>                                      \
+        struct AlignBytes<t>                            \
+        {                                               \
+            enum                                        \
+            {                                           \
+                ALIGN_BYTES = b                         \
+            };                                          \
+            using Type = __attribute__((aligned(b))) t; \
+        };
 
 __HIPCUB_ALIGN_BYTES(short4, 8)
 __HIPCUB_ALIGN_BYTES(ushort4, 8)
@@ -300,28 +302,26 @@ struct UnitWord
     };
 
     /// Biggest shuffle word that T is a whole multiple of and is not larger than the alignment of T
-    typedef typename std::conditional<IsMultiple<int>::IS_MULTIPLE,
-        unsigned int,
-        typename std::conditional<IsMultiple<short>::IS_MULTIPLE,
-            unsigned short,
-            unsigned char>::type>::type         ShuffleWord;
+    using ShuffleWord =
+        typename std::conditional<IsMultiple<int>::IS_MULTIPLE,
+                                  unsigned int,
+                                  typename std::conditional<IsMultiple<short>::IS_MULTIPLE,
+                                                            unsigned short,
+                                                            unsigned char>::type>::type;
 
     /// Biggest volatile word that T is a whole multiple of and is not larger than the alignment of T
-    typedef typename std::conditional<IsMultiple<long long>::IS_MULTIPLE,
-        unsigned long long,
-        ShuffleWord>::type                      VolatileWord;
+    using VolatileWord = typename std::
+        conditional<IsMultiple<long long>::IS_MULTIPLE, unsigned long long, ShuffleWord>::type;
 
     /// Biggest memory-access word that T is a whole multiple of and is not larger than the alignment of T
-    typedef typename std::conditional<IsMultiple<longlong2>::IS_MULTIPLE,
-        ulonglong2,
-        VolatileWord>::type                     DeviceWord;
+    using DeviceWord = typename std::
+        conditional<IsMultiple<longlong2>::IS_MULTIPLE, ulonglong2, VolatileWord>::type;
 
     /// Biggest texture reference word that T is a whole multiple of and is not larger than the alignment of T
-    typedef typename std::conditional<IsMultiple<int4>::IS_MULTIPLE,
+    using TextureWord = typename std::conditional<
+        IsMultiple<int4>::IS_MULTIPLE,
         uint4,
-        typename std::conditional<IsMultiple<int2>::IS_MULTIPLE,
-            uint2,
-            ShuffleWord>::type>::type           TextureWord;
+        typename std::conditional<IsMultiple<int2>::IS_MULTIPLE, uint2, ShuffleWord>::type>::type;
 };
 
 
@@ -329,20 +329,20 @@ struct UnitWord
 template <>
 struct UnitWord <float2>
 {
-    typedef int         ShuffleWord;
-    typedef unsigned long long   VolatileWord;
-    typedef unsigned long long   DeviceWord;
-    typedef float2      TextureWord;
+    using ShuffleWord  = int;
+    using VolatileWord = unsigned long long;
+    using DeviceWord   = unsigned long long;
+    using TextureWord  = float2;
 };
 
 // float4 specialization workaround (for SM10-SM13)
 template <>
 struct UnitWord <float4>
 {
-    typedef int         ShuffleWord;
-    typedef unsigned long long  VolatileWord;
-    typedef ulonglong2          DeviceWord;
-    typedef float4              TextureWord;
+    using ShuffleWord  = int;
+    using VolatileWord = unsigned long long;
+    using DeviceWord   = ulonglong2;
+    using TextureWord  = float4;
 };
 
 
@@ -350,10 +350,10 @@ struct UnitWord <float4>
 template <>
 struct UnitWord <char2>
 {
-    typedef unsigned short      ShuffleWord;
-    typedef unsigned short      VolatileWord;
-    typedef unsigned short      DeviceWord;
-    typedef unsigned short      TextureWord;
+    using ShuffleWord  = unsigned short;
+    using VolatileWord = unsigned short;
+    using DeviceWord   = unsigned short;
+    using TextureWord  = unsigned short;
 };
 
 
@@ -361,11 +361,98 @@ template <typename T> struct UnitWord<volatile T> : UnitWord<T> {};
 template <typename T> struct UnitWord<const T> : UnitWord<T> {};
 template <typename T> struct UnitWord<const volatile T> : UnitWord<T> {};
 
-
 #endif // DOXYGEN_SHOULD_SKIP_THIS
 
+/******************************************************************************
+ * Vector type inference utilities.
+ ******************************************************************************/
 
+template<typename T, int vec_elements>
+struct GenericCubVector
+{
+    static_assert(!sizeof(T), "CubVector can only have 1-4 elements");
+};
 
+enum
+{
+    /// The maximum number of elements in HIP vector types
+    MAX_VEC_ELEMENTS = 4,
+};
+
+template<typename T>
+struct GenericCubVector<T, 1>
+{
+    T x;
+
+    using BaseType = T;
+    using Type     = GenericCubVector<T, 1>;
+};
+
+template<typename T>
+struct GenericCubVector<T, 2>
+{
+    T x;
+    T y;
+
+    using BaseType = T;
+    using Type     = GenericCubVector<T, 2>;
+};
+
+template<typename T>
+struct GenericCubVector<T, 3>
+{
+    T x;
+    T y;
+    T z;
+
+    using BaseType = T;
+    using Type     = GenericCubVector<T, 3>;
+};
+
+template<typename T>
+struct GenericCubVector<T, 4>
+{
+    T x;
+    T y;
+    T z;
+    T w;
+
+    using BaseType = T;
+    using Type     = GenericCubVector<T, 4>;
+};
+
+template<typename T, int vec_elements>
+struct CubVectorType
+{
+    // Fallback on GenericCubVector
+    using Type = GenericCubVector<T, vec_elements>;
+};
+
+template<typename T, int vec_elements>
+using CubVector = typename CubVectorType<T, vec_elements>::Type;
+
+#define HIPCUB_DEFINE_VECTOR_TYPE(base_type, vec_type)        \
+    template<int vec_elements>                                \
+    struct CubVectorType<base_type, vec_elements>             \
+    {                                                         \
+        using Type = HIP_vector_type<vec_type, vec_elements>; \
+    };
+
+HIPCUB_DEFINE_VECTOR_TYPE(char, char)
+HIPCUB_DEFINE_VECTOR_TYPE(unsigned char, unsigned char)
+HIPCUB_DEFINE_VECTOR_TYPE(short, short)
+HIPCUB_DEFINE_VECTOR_TYPE(ushort, ushort)
+HIPCUB_DEFINE_VECTOR_TYPE(int, int)
+HIPCUB_DEFINE_VECTOR_TYPE(uint, uint)
+HIPCUB_DEFINE_VECTOR_TYPE(long, long)
+HIPCUB_DEFINE_VECTOR_TYPE(unsigned long, unsigned long)
+HIPCUB_DEFINE_VECTOR_TYPE(long long, long long)
+HIPCUB_DEFINE_VECTOR_TYPE(unsigned long long, unsigned long long)
+HIPCUB_DEFINE_VECTOR_TYPE(float, float)
+HIPCUB_DEFINE_VECTOR_TYPE(double, double)
+HIPCUB_DEFINE_VECTOR_TYPE(bool, unsigned char)
+
+#undef HIPCUB_DEFINE_VECTOR_TYPE
 
 /******************************************************************************
  * Wrapper types
@@ -378,7 +465,7 @@ template <typename T>
 struct Uninitialized
 {
     /// Biggest memory-access word that T is a whole multiple of and is not larger than the alignment of T
-    typedef typename UnitWord<T>::DeviceWord DeviceWord;
+    using DeviceWord = typename UnitWord<T>::DeviceWord;
 
     static constexpr std::size_t DATA_SIZE = sizeof(T);
     static constexpr std::size_t WORD_SIZE = sizeof(DeviceWord);
@@ -394,24 +481,18 @@ struct Uninitialized
     }
 };
 
-
 /******************************************************************************
  * Simple type traits utilities.
- *
- * For example:
- *     Traits<int>::CATEGORY             // SIGNED_INTEGER
- *     Traits<NullType>::NULL_TYPE       // true
- *     Traits<uint4>::CATEGORY           // NOT_A_NUMBER
- *     Traits<uint4>::PRIMITIVE;         // false
- *
  ******************************************************************************/
 
- #ifndef DOXYGEN_SHOULD_SKIP_THIS    // Do not document
+#ifndef DOXYGEN_SHOULD_SKIP_THIS // Do not document
 
 /**
- * \brief Basic type traits categories
+ * \brief Basic type traits categories.
+ * This enum is deprecated, please use <rocprim/type_traits> instead. Or if you have
+ * libhipcxx, please use the type_traits system in libhipcxx.
  */
-enum Category
+enum HIPCUB_DEPRECATED_BECAUSE("Use <rocprim/type_traits> instead.") Category
 {
     NOT_A_NUMBER,
     SIGNED_INTEGER,
@@ -419,41 +500,49 @@ enum Category
     FLOATING_POINT
 };
 
-
 /**
  * \brief Basic type traits
  */
-template <Category _CATEGORY, bool _PRIMITIVE, bool _NULL_TYPE, typename _UnsignedBits, typename T>
+HIPCUB_CLANG_SUPPRESS_DEPRECATED_PUSH
+template<Category _CATEGORY,
+         bool     _PRIMITIVE,
+         bool     _nullptr_TYPE,
+         typename _UnsignedBits,
+         typename T>
 struct BaseTraits
 {
     /// Category
-    static const Category CATEGORY      = _CATEGORY;
+    HIPCUB_DEPRECATED_BECAUSE("Use <rocprim/type_traits> instead.")
+    static constexpr Category CATEGORY = _CATEGORY;
     enum
     {
-        PRIMITIVE       = _PRIMITIVE,
-        NULL_TYPE       = _NULL_TYPE,
+        PRIMITIVE HIPCUB_DEPRECATED_BECAUSE("Use <rocprim/type_traits> instead.") = _PRIMITIVE,
+        nullptr_TYPE                                                              = _nullptr_TYPE,
     };
 };
-
+HIPCUB_CLANG_SUPPRESS_DEPRECATED_POP
 
 /**
  * Basic type traits (unsigned primitive specialization)
  */
+HIPCUB_CLANG_SUPPRESS_DEPRECATED_PUSH
 template <typename _UnsignedBits, typename T>
 struct BaseTraits<UNSIGNED_INTEGER, true, false, _UnsignedBits, T>
 {
-    typedef _UnsignedBits       UnsignedBits;
+    using UnsignedBits = _UnsignedBits;
 
-    static const Category       CATEGORY    = UNSIGNED_INTEGER;
-    static const UnsignedBits   LOWEST_KEY  = UnsignedBits(0);
-    static const UnsignedBits   MAX_KEY     = UnsignedBits(-1);
+    HIPCUB_DEPRECATED_BECAUSE("Use <rocprim/type_traits> instead.")
+    static constexpr Category     CATEGORY   = UNSIGNED_INTEGER;
+    static constexpr UnsignedBits LOWEST_KEY = UnsignedBits(0);
+    static constexpr UnsignedBits MAX_KEY    = UnsignedBits(-1);
 
     enum
     {
-        PRIMITIVE       = true,
-        NULL_TYPE       = false,
+        PRIMITIVE HIPCUB_DEPRECATED_BECAUSE("Use <rocprim/type_traits> instead.") = true,
+        nullptr_TYPE                                                              = false,
     };
 
+    using key_codec = decltype(::rocprim::traits::get<T>().template radix_key_codec<false>());
 
     static HIPCUB_HOST_DEVICE __forceinline__ UnsignedBits TwiddleIn(UnsignedBits key)
     {
@@ -481,27 +570,31 @@ struct BaseTraits<UNSIGNED_INTEGER, true, false, _UnsignedBits, T>
         return retval;
     }
 };
-
+HIPCUB_CLANG_SUPPRESS_DEPRECATED_POP
 
 /**
  * Basic type traits (signed primitive specialization)
  */
+HIPCUB_CLANG_SUPPRESS_DEPRECATED_PUSH
 template <typename _UnsignedBits, typename T>
 struct BaseTraits<SIGNED_INTEGER, true, false, _UnsignedBits, T>
 {
-    typedef _UnsignedBits       UnsignedBits;
+    using UnsignedBits = _UnsignedBits;
 
-    static const Category       CATEGORY    = SIGNED_INTEGER;
-    static const UnsignedBits   HIGH_BIT    = UnsignedBits(1) << ((sizeof(UnsignedBits) * 8) - 1);
-    static const UnsignedBits   LOWEST_KEY  = HIGH_BIT;
-    static const UnsignedBits   MAX_KEY     = UnsignedBits(-1) ^ HIGH_BIT;
+    HIPCUB_DEPRECATED_BECAUSE("Use <rocprim/type_traits> instead.")
+    static constexpr Category     CATEGORY   = SIGNED_INTEGER;
+    static constexpr UnsignedBits HIGH_BIT   = UnsignedBits(1) << ((sizeof(UnsignedBits) * 8) - 1);
+    static constexpr UnsignedBits LOWEST_KEY = HIGH_BIT;
+    static constexpr UnsignedBits MAX_KEY    = UnsignedBits(-1) ^ HIGH_BIT;
 
     enum
     {
-        PRIMITIVE       = true,
-        NULL_TYPE       = false,
+        PRIMITIVE HIPCUB_DEPRECATED_BECAUSE("Use <rocprim/type_traits> instead.") = true,
+        nullptr_TYPE                                                              = false,
     };
 
+    using key_codec = decltype(::rocprim::traits::get<T>().template radix_key_codec<false>());
+
     static HIPCUB_HOST_DEVICE __forceinline__ UnsignedBits TwiddleIn(UnsignedBits key)
     {
         return key ^ HIGH_BIT;
@@ -524,10 +617,13 @@ struct BaseTraits<SIGNED_INTEGER, true, false, _UnsignedBits, T>
         return reinterpret_cast<T&>(retval);
     }
 };
+HIPCUB_CLANG_SUPPRESS_DEPRECATED_POP
 
+// This API needs to be deprecated once libhipcxx is available.
 template <typename _T>
 struct FpLimits;
 
+// This API needs to be deprecated once libhipcxx is available.
 template <>
 struct FpLimits<float>
 {
@@ -540,6 +636,7 @@ struct FpLimits<float>
     }
 };
 
+// This API needs to be deprecated once libhipcxx is available.
 template <>
 struct FpLimits<double>
 {
@@ -552,6 +649,7 @@ struct FpLimits<double>
     }
 };
 
+// This API needs to be deprecated once libhipcxx is available.
 template <>
 struct FpLimits<__half>
 {
@@ -566,6 +664,7 @@ struct FpLimits<__half>
     }
 };
 
+// This API needs to be deprecated once libhipcxx is available.
 template <>
 struct FpLimits<hip_bfloat16>
 {
@@ -583,20 +682,24 @@ struct FpLimits<hip_bfloat16>
 /**
  * Basic type traits (fp primitive specialization)
  */
+HIPCUB_CLANG_SUPPRESS_DEPRECATED_PUSH
 template <typename _UnsignedBits, typename T>
 struct BaseTraits<FLOATING_POINT, true, false, _UnsignedBits, T>
 {
-    typedef _UnsignedBits       UnsignedBits;
+    using UnsignedBits = _UnsignedBits;
 
-    static const Category       CATEGORY    = FLOATING_POINT;
-    static const UnsignedBits   HIGH_BIT    = UnsignedBits(1) << ((sizeof(UnsignedBits) * 8) - 1);
-    static const UnsignedBits   LOWEST_KEY  = UnsignedBits(-1);
-    static const UnsignedBits   MAX_KEY     = UnsignedBits(-1) ^ HIGH_BIT;
+    HIPCUB_DEPRECATED_BECAUSE("Use <rocprim/type_traits> instead.")
+    static constexpr Category     CATEGORY   = FLOATING_POINT;
+    static constexpr UnsignedBits HIGH_BIT   = UnsignedBits(1) << ((sizeof(UnsignedBits) * 8) - 1);
+    static constexpr UnsignedBits LOWEST_KEY = UnsignedBits(-1);
+    static constexpr UnsignedBits MAX_KEY    = UnsignedBits(-1) ^ HIGH_BIT;
+
+    using key_codec = decltype(::rocprim::traits::get<T>().template radix_key_codec<false>());
 
     enum
     {
-        PRIMITIVE       = true,
-        NULL_TYPE       = false,
+        PRIMITIVE HIPCUB_DEPRECATED_BECAUSE("Use <rocprim/type_traits> instead.") = true,
+        nullptr_TYPE                                                              = false,
     };
 
     static HIPCUB_HOST_DEVICE __forceinline__ UnsignedBits TwiddleIn(UnsignedBits key)
@@ -619,11 +722,12 @@ struct BaseTraits<FLOATING_POINT, true, false, _UnsignedBits, T>
         return FpLimits<T>::Lowest();
     }
 };
-
+HIPCUB_CLANG_SUPPRESS_DEPRECATED_POP
 
 /**
  * \brief Numeric type traits
  */
+HIPCUB_CLANG_SUPPRESS_DEPRECATED_PUSH
 template <typename T> struct NumericTraits :            BaseTraits<NOT_A_NUMBER, false, false, T, T> {};
 
 template <> struct NumericTraits<NullType> :            BaseTraits<NOT_A_NUMBER, false, true, NullType, NullType> {};
@@ -652,8 +756,11 @@ struct NumericTraits<__uint128_t>
     static constexpr UnsignedBits LOWEST_KEY = UnsignedBits(0);
     static constexpr UnsignedBits MAX_KEY    = UnsignedBits(-1);
 
-    static constexpr bool PRIMITIVE = false;
-    static constexpr bool NULL_TYPE = false;
+    HIPCUB_DEPRECATED_BECAUSE("Use <rocprim/type_traits> instead.")
+    static constexpr bool PRIMITIVE    = false;
+    static constexpr bool nullptr_TYPE = false;
+
+    using key_codec = decltype(::rocprim::traits::get<T>().template radix_key_codec<false>());
 
     static __host__ __device__ __forceinline__ UnsignedBits TwiddleIn(UnsignedBits key)
     {
@@ -687,8 +794,11 @@ struct NumericTraits<__int128_t>
     static constexpr UnsignedBits LOWEST_KEY = HIGH_BIT;
     static constexpr UnsignedBits MAX_KEY    = UnsignedBits(-1) ^ HIGH_BIT;
 
+    HIPCUB_DEPRECATED_BECAUSE("Use <rocprim/type_traits> instead.")
     static constexpr bool PRIMITIVE = false;
-    static constexpr bool NULL_TYPE = false;
+    static constexpr bool nullptr_TYPE = false;
+
+    using key_codec = decltype(::rocprim::traits::get<T>().template radix_key_codec<false>());
 
     static __host__ __device__ __forceinline__ UnsignedBits TwiddleIn(UnsignedBits key)
     {
@@ -720,15 +830,195 @@ template <> struct NumericTraits<__half> :              BaseTraits<FLOATING_POIN
 template <> struct NumericTraits<hip_bfloat16 > :       BaseTraits<FLOATING_POINT, true, false, unsigned short, hip_bfloat16 > {};
 
 template <> struct NumericTraits<bool> :                BaseTraits<UNSIGNED_INTEGER, true, false, typename UnitWord<bool>::VolatileWord, bool> {};
+HIPCUB_CLANG_SUPPRESS_DEPRECATED_POP
 
 /**
  * \brief Type traits
  */
-template <typename T>
-struct Traits : NumericTraits<typename std::remove_cv<T>::type> {};
+template<typename T>
+struct Traits : NumericTraits<typename std::remove_cv<T>::type>
+{};
+
+namespace detail
+{
+// __uint128_t and __int128_t are not primitive
+
+HIPCUB_CLANG_SUPPRESS_DEPRECATED_PUSH
+template<typename T>
+struct is_primitive : ::std::bool_constant<Traits<T>::PRIMITIVE>
+{};
+
+template<typename T>
+inline constexpr bool is_primitive_v = is_primitive<T>::value;
+
+HIPCUB_CLANG_SUPPRESS_DEPRECATED_POP
+
+} // namespace detail
+
+/**
+ * \brief Common type of zero types.
+ */
+template<class...>
+struct common_type
+{};
+
+/**
+ * \brief Common type of a single type.
+ */
+template<class T>
+struct common_type<T> : common_type<T, T>
+{};
+
+// Common type of a pair of types.
+namespace detail
+{
+
+// Determines if type is half or bfloat16 (extended fp).
+template<class T>
+struct is_extended_fp
+    : std::integral_constant<
+          bool,
+          std::is_same<__half, typename std::remove_cv<T>::type>::value
+              || std::is_same<hip_bfloat16, typename std::remove_cv<T>::type>::value>
+{};
+
+// Gets "raw" type: drops reference and const qualifier.
+template<typename T>
+struct remove_cvref
+{
+    using type = typename std::remove_cv<typename std::remove_reference<T>::type>::type;
+};
+template<typename T>
+using remove_cvref_t = typename remove_cvref<T>::type;
+
+template<template<typename...> class MFn, bool condition, typename T>
+using apply_if_t = std::conditional_t<condition, MFn<T>, T>;
+
+template<typename From, typename To>
+using copy_cv_t
+    = apply_if_t<std::add_volatile_t,
+                 std::is_volatile_v<From>,
+                 apply_if_t<std::add_const_t, std::is_const_v<From>, std::remove_cv_t<To>>>;
+
+template<typename From, typename To>
+using copy_ref_t = apply_if_t<std::add_rvalue_reference_t,
+                              std::is_rvalue_reference_v<From>,
+                              apply_if_t<std::add_lvalue_reference_t,
+                                         std::is_lvalue_reference_v<From>,
+                                         std::remove_reference_t<To>>>;
+
+template<typename From, typename To>
+using copy_cvref_t = copy_ref_t<From, copy_cv_t<std::remove_reference_t<From>, remove_cvref_t<To>>>;
+
+// Captures non extended fp types.
+template<class T1, class T2, class = void>
+struct common_type_extended_fp
+{
+    using type = typename std::common_type<T1, T2>::type;
+};
+
+// Captures first type arithmetic, second one extended FP.
+template<class T1, class T2>
+struct common_type_extended_fp<
+    T1,
+    T2,
+    typename std::enable_if_t<std::is_arithmetic<remove_cvref_t<T1>>::value
+                              && is_extended_fp<remove_cvref_t<T2>>::value>>
+{
+    using type = typename std::common_type<T1, copy_cvref_t<T2, float>>::type;
+};
+
+// Captures first type extended FP, second one arithmetic.
+template<class T1, class T2>
+struct common_type_extended_fp<
+    T1,
+    T2,
+    typename std::enable_if_t<is_extended_fp<remove_cvref_t<T1>>::value
+                              && std::is_arithmetic<remove_cvref_t<T2>>::value>>
+{
+    using type = typename std::common_type<copy_cvref_t<T1, float>, T2>::type;
+};
+
+template<class...>
+using void_t = void;
+
+template<int Value>
+using int_constant_t = ::std::integral_constant<int, Value>;
+
+// Common type of more than two types.
+
+template<class AlwaysVoid, class T1, class T2, class... Rest>
+struct common_type_multi_impl
+{};
+
+template<class T1, class T2, class... Rest>
+struct common_type_multi_impl<void_t<typename common_type<T1, T2>::type>, T1, T2, Rest...>
+    : common_type<typename common_type<T1, T2>::type, Rest...>
+{};
+} // namespace detail
+
+/**
+ * \brief Common type of a pair of types
+ */
+template<class T1, class T2>
+struct common_type<T1, T2> : detail::common_type_extended_fp<T1, T2>
+{};
+
+/**
+ * \brief Common type of more than two types
+ */
+template<class T1, class T2, class... Rest>
+struct common_type<T1, T2, Rest...> : detail::common_type_multi_impl<void, T1, T2, Rest...>
+{};
+
+/**
+ * \brief Common type helper
+ */
+template<class... Ts>
+using common_type_t = typename common_type<Ts...>::type;
 
 #endif // DOXYGEN_SHOULD_SKIP_THIS
 
+namespace detail
+{
+
+template<typename T, typename = void>
+struct is_fixed_size_random_access_range : ::std::false_type
+{};
+
+template<typename T, size_t N>
+struct is_fixed_size_random_access_range<T[N], void> : ::std::true_type
+{};
+
+template<typename T>
+struct is_fixed_size_random_access_range<T, void_t<decltype(std::declval<T&>()[0])>>
+    : ::std::true_type
+{};
+
+// TODO: for is_fixed_size_random_access_range we also need to support array span and extents after we can use libhipcxx.
+template<typename T>
+using is_fixed_size_random_access_range_t = typename is_fixed_size_random_access_range<T>::type;
+
+template<typename T, typename = void>
+struct static_size
+{
+    static_assert(false, "static_size not supported for this type");
+};
+
+template<typename T, size_t N>
+struct static_size<T[N], void> : ::std::integral_constant<int, N>
+{};
+
+template<typename T>
+[[nodiscard]]
+__host__ __device__ __forceinline__
+constexpr size_t static_size_v()
+{
+    return static_size<T>::value;
+}
+
+} // namespace detail
+
 END_HIPCUB_NAMESPACE
 
 #endif // HIPCUB_ROCPRIM_UTIL_TYPE_HPP_
