@@ -1,7 +1,7 @@
 /******************************************************************************
  * Copyright (c) 2010-2011, Duane Merrill.  All rights reserved.
  * Copyright (c) 2011-2018, NVIDIA CORPORATION.  All rights reserved.
- * Modifications Copyright (c) 2019-2024, Advanced Micro Devices, Inc.  All rights reserved.
+ * Modifications Copyright (c) 2019-2025, Advanced Micro Devices, Inc.  All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions are met:
@@ -32,19 +32,52 @@
 
 #include <hip/hip_runtime.h>
 
+// Version
+#include "hipcub_version.hpp" // IWYU pragma: export
+
 #define HIPCUB_NAMESPACE hipcub
 
-#define BEGIN_HIPCUB_NAMESPACE \
-    namespace hipcub {
+// Inline namespace (e.g. HIPCUB_300400_NS where 300400 is the hipCUB version) is used to
+// eliminate issues when shared libraries are built with different versions of hipCUB so they may
+// have symbols with the same name but different content.
+// HIPCUB_DISABLE_INLINE_NAMESPACE can be defined to disable inline namespaces (the old behavior).
+// HIPCUB_INLINE_NAMESPACE can be defined to override the standard inline namespace name.
+// Additionally, all kernels have hidden visibility.
+// For rocPRIM backend, see rocprim/config.hpp for similar definitions.
+#if defined(DOXYGEN_SHOULD_SKIP_THIS) || defined(HIPCUB_DISABLE_INLINE_NAMESPACE)
+    #define HIPCUB_INLINE_NAMESPACE
+    #define BEGIN_HIPCUB_INLINE_NAMESPACE
+    #define END_HIPCUB_INLINE_NAMESPACE
+#else
+    #define HIPCUB_CONCAT_(SEP, A, B) A##SEP##B
+    #define HIPCUB_CONCAT(SEP, A, B) HIPCUB_CONCAT_(SEP, A, B)
 
-#define END_HIPCUB_NAMESPACE \
-    } /* hipcub */
+    #ifndef HIPCUB_INLINE_NAMESPACE
+        #define HIPCUB_INLINE_NAMESPACE \
+            HIPCUB_CONCAT(_, HIPCUB, HIPCUB_CONCAT(_, HIPCUB_VERSION, NS))
+    #endif
+    #define BEGIN_HIPCUB_INLINE_NAMESPACE        \
+        inline namespace HIPCUB_INLINE_NAMESPACE \
+        {
+    #define END_HIPCUB_INLINE_NAMESPACE } /* inline namespace */
+#endif
+
+#define BEGIN_HIPCUB_NAMESPACE \
+    namespace HIPCUB_NAMESPACE \
+    {                          \
+    BEGIN_HIPCUB_INLINE_NAMESPACE
+
+#define END_HIPCUB_NAMESPACE    \
+    END_HIPCUB_INLINE_NAMESPACE \
+    } /* namespace hipcub */
 
 #ifdef __HIP_PLATFORM_AMD__
     #define HIPCUB_ROCPRIM_API 1
     #define HIPCUB_RUNTIME_FUNCTION __host__
 
     #include <rocprim/device/config_types.hpp>
+    #include <rocprim/intrinsics/arch.hpp>
+    #include <rocprim/intrinsics/thread.hpp>
 
 BEGIN_HIPCUB_NAMESPACE
 namespace detail
@@ -67,9 +100,11 @@ inline unsigned int host_warp_size_wrapper()
 }
 } // namespace detail
 END_HIPCUB_NAMESPACE
+    #include <rocprim/intrinsics/arch.hpp>
 
     #define HIPCUB_WARP_THREADS ::rocprim::warp_size()
-    #define HIPCUB_DEVICE_WARP_THREADS ::rocprim::device_warp_size()
+    // HIPCUB (and CUB) don't have a method to express min and max warp size.
+    #define HIPCUB_DEVICE_WARP_THREADS ::rocprim::arch::wavefront::max_size()
     #define HIPCUB_HOST_WARP_THREADS ::hipcub::detail::host_warp_size_wrapper()
     #define HIPCUB_ARCH 1 // ignored with rocPRIM backend
 #elif defined(__HIP_PLATFORM_NVIDIA__)
